﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;

namespace CPF_experement
{
    class ECBSE : CBSE
    {

        public ECBSE(IDnCSolver solver, double eps, int maxThreshold=-1, int currentThreshold=-1): base(solver,maxThreshold,currentThreshold)
        {
            this.focalList = new BinaryHeap();
            this.eps = eps;
        }

        public override void Setup(ProblemInstance problemInstance)
        {
            LBCheck = true;
            this.instance = problemInstance;
            DnCNode root = new DnCNode(instance.m_vAgents.Length);
            this.openList.Add(root);
            this.highLevelExpanded = 0;
            this.highLevelGenerated = 1;
            loweLevelExpanded = 0;
            loweLevelGenerated = 0;
            maxSizeGroup = 1;
            this.totalCost = 0;

            this.focalList.Add(new FocalNodeCBS(root));

            if (problemInstance.parameters.ContainsKey(Trevor.MAXIMUM_COST_KEY))
                this.maxCost = (int)(problemInstance.parameters[Trevor.MAXIMUM_COST_KEY]);
            else
                this.maxCost = int.MaxValue;
            if (problemInstance.parameters.ContainsKey(CBS_LocalConflicts.NEW_INTERNAL_CAT) == false)
            {
                problemInstance.parameters[CBS_LocalConflicts.INTERNAL_CAT] = new HashSet_U<TimedMove>();
                problemInstance.parameters[CBS_LocalConflicts.CONSTRAINTS] = new HashSet_U<DnCConstraint>();
                problemInstance.parameters[CBS_LocalConflicts.NEW_CONSTRAINTS] = new HashSet<DnCConstraint>();
                problemInstance.parameters[CBS_LocalConflicts.NEW_INTERNAL_CAT] = new HashSet<TimedMove>();
            }
            else
            {
                problemInstance.parameters[CBS_LocalConflicts.NEW_INTERNAL_CAT] = new HashSet<TimedMove>();
                problemInstance.parameters[CBS_LocalConflicts.NEW_CONSTRAINTS] = new HashSet<DnCConstraint>();
            }
            DnCNode.allConstraintsForNode = new HashSet<DnCConstraint>();
            minCost = 0;
            isGlobal = false;
            LBCheck = false;
        }

        public override String GetName()
        {
            if (mergeThreshold == -1)
                return "ECBSE"+eps+"-"+solver.GetName();
            return "ECBSE Local(" + mergeThreshold + ")(" + maxThreshold + ")+" + lowLevelSolver.GetName();
        }

        public override bool Solve(Run runner)
        {
            //Debug.WriteLine("Solving Sub-problem On Level - " + mergeThreshold);
            //Console.ReadLine();

            isDnC = true;
            DnCConflict conflict;

            FocalNodeCBS next;
            DnCNode temp;
            int newF;
            int newThreshold;

            this.runner = runner;

            next = (FocalNodeCBS)focalList.Remove();
            DnCNode currentNode = next.orgnl;
            openList.Remove(currentNode);

            long lastTime = -1;
            LBCheck = true;

            //Console.WriteLine("TC|CA|VC|COU|CON|CEX");

            highLevelExpanded++;
            if (currentNode.solve(instance, runner, minCost, lowLevelSolver, ref highLevelExpanded, ref highLevelGenerated, ref loweLevelExpanded, ref loweLevelGenerated) == false)
            {
                return false;
            }

            //ECBSE Change
            fMin = currentNode.totalLB;
            threshold = (int)Math.Floor(fMin * eps);

            if (currentNode.totalCost <= this.maxCost)
            {
                this.openList.Add(currentNode);
                this.closedList.Add(currentNode);
                this.focalList.Add(new FocalNodeCBS(currentNode));
                this.addToGlobalConflictCount(currentNode.getConflict());
            }
            while (focalList.Count > 0 && runner.ElapsedMilliseconds() < Constants.MAX_TIME)
            {
                // Check if max time has been exceeded
                if (runner.ElapsedMilliseconds() > Constants.MAX_TIME)
                {
                    totalCost = Constants.TIMEOUT_COST;
                    Console.WriteLine("Out of time");
                    isDnC = false;
                    this.Clear();
                    return false;
                }

                next = (FocalNodeCBS)focalList.Remove();
                currentNode = next.orgnl;
                openList.Remove(currentNode);


                //Console.Write("Depth: " + currentNode.depth + ",");
                //Console.Write("TC: " + currentNode.totalCost + ",");
                //Console.Write("Conflict: " + currentNode.internalConflictsCount + ",");
                //Console.Write("Couple: " + currentNode.couple + ",");
                //Console.Write("Replan: " + currentNode.replanNum + ",");
                //Console.WriteLine("VCover: " + currentNode.vCover);

                conflict = currentNode.getConflict();
                // Check if node is the goal
                if (conflict == null)
                {
                    this.totalCost = currentNode.totalCost;
                    this.goalNode = currentNode;
                    this.solution = currentNode.CalculateJointPlan();
                    isDnC = false;
                    this.Clear();

                    Console.WriteLine("MinLB: " + fMin + " Sol LB: " + currentNode.totalLB);

                    //TextWriter resultsWriter = new StreamWriter("Depth.csv", true);
                    //resultsWriter.WriteLine(currentNode.depth);
                    //resultsWriter.Flush();
                    //resultsWriter.Close();
                    //Console.WriteLine("EX: " + currentNode.totalCost + "|" + currentNode.replanNum + "|" + currentNode.vCover + "|" + currentNode.couple + "|" + currentNode.internalConflictsCount + "|" + currentNode.costExp);

                    //foreach (SinglePlan plan in currentNode.allSingleAgentPlans)
                    //{
                    //    plan.PrintPlan();
                    //}

                    return true;
                }
                // Expand
                highLevelExpanded++;

                //if (runner.ElapsedMilliseconds() / 60000 > lastTime)
                //{
                //    Console.WriteLine("EX: " + currentNode.totalCost + "|" + currentNode.replanNum + "|" + currentNode.vCover + "|" + currentNode.couple + "|" + currentNode.internalConflictsCount + "|" + currentNode.costExp);
                //    Console.WriteLine(runner.ElapsedMilliseconds() + "|" +highLevelGenerated + "|" + highLevelExpanded + "|" + loweLevelGenerated + "|" + loweLevelExpanded + "|" + currentNode.depth);
                //    lastTime = runner.ElapsedMilliseconds() / 60000;
                //}
                
                if (expand(currentNode, conflict))
                    if (currentNode.colapse == 0)
                        currentNode.clear();

                //check if fMin changed
                temp = (DnCNode)openList.Peek();
                Debug.Assert(temp != null);

                newF = temp.totalLB;
                if (newF > fMin)
                {
                    fMin = newF;
                    newThreshold = (int)Math.Floor(fMin * eps);
                    updateFocal(threshold, newThreshold);
                    threshold = newThreshold;
                }
            }
            totalCost = Constants.NO_SOLUTION_COST;
            isDnC = false;
            LBCheck = false;
            this.Clear();
            return false;
        }   

        public override bool expand(DnCNode node, DnCConflict conflict)
        {
            closedList.Remove(node);

            //TODO: Not implemented CBSE for MA-CBS
            if (this.maxThreshold != -1 && checkMerge(node))
            {
                if (closedList.Contains(node))
                    return true;
                if (node.rePlan(instance, runner, conflict.agentA, this.minCost, solver, lowLevelSolver, ref highLevelExpanded, ref highLevelGenerated, ref loweLevelExpanded, ref loweLevelGenerated) == false)
                {
                    if (node.replanSize > this.maxSizeGroup)
                        maxSizeGroup = node.replanSize;
                    return true;
                }
                if (node.replanSize > this.maxSizeGroup)
                    maxSizeGroup = node.replanSize;
                if (node.totalCost <= maxCost)
                    openList.Add(node);
                closedList.Add(node);
                this.addToGlobalConflictCount(node.getConflict());
                return false;
            }

            closedList.Add(node);

            int prtCost = node.totalCost;
            DnCConstraint con;
            DnCNode toAdd;

            //Console.WriteLine("EX: " + node.totalCost + "|" + node.replanNum + "|" + node.vCover + "|" + node.couple + "|" + node.internalConflictsCount + "|" + node.costExp);

            //foreach (SinglePlan plan in node.allSingleAgentPlans)
            //{
            //    plan.PrintPlan();
            //}

            bool colapse = false;
            if (node.colapse != 0)
                colapse = true;

            //colapse = 1 - 1st wasnot expanded 
            //colapse = 2 - 2st wasnot expanded 
            //colapse = 0 - both not expanded 
            if (node.colapse != 1 && Math.Max(minCost, conflict.timeStep) > node.pathLength(conflict.agentA))
            {
                node.colapse = 1;
                node.totalCost += (conflict.timeStep + 1 - node.pathLength(conflict.agentA));
                node.totalLB += (conflict.timeStep + 1 - node.pathLength(conflict.agentA));
                openList.Add(node);
                if (node.totalCost <= threshold)
                    this.focalList.Add(new FocalNodeCBS(node));

                //Console.WriteLine("Colapse");
            }
            else if (!colapse || node.colapse == 1)
            {
                con = new DnCConstraint(conflict, instance, true);
                toAdd = new DnCNode(node, con, conflict.agentA, instance);

                if (closedList.Contains(toAdd) == false)
                {

                    if (toAdd.rePlan(instance, runner, conflict.agentA, Math.Max(minCost, getReplanDepth(node, conflict.agentA)), solver, lowLevelSolver, ref highLevelExpanded, ref highLevelGenerated, ref loweLevelExpanded, ref loweLevelGenerated))
                    {
                        if (toAdd.totalCost <= this.maxCost)
                        {
                            //if (toAdd.totalCost < prtCost)
                            //    Console.WriteLine("Stop!!");
                            
                            //if (toAdd.totalCost < fMin)
                            //    Console.WriteLine("Stop!!");

                            //if (toAdd.totalCost <= fMin / 1.1)
                            //    Console.WriteLine("Stop!!");

                            //Console.WriteLine("GEN: " + toAdd.totalCost + "|" + toAdd.replanNum + "|" + toAdd.vCover + "|" + toAdd.couple + "|" + toAdd.internalConflictsCount + "|" + toAdd.costExp);
                            //foreach (SinglePlan plan in toAdd.allSingleAgentPlans)
                            //{
                            //    plan.PrintPlan();
                            //}

                            openList.Add(toAdd);
                            closedList.Add(toAdd);
                            this.highLevelGenerated++;
                            addToGlobalConflictCount(toAdd.getConflict());

                            if (toAdd.totalCost <= threshold)
                                this.focalList.Add(new FocalNodeCBS(toAdd));
                        }
                    }
                }
                //else
                //    Console.WriteLine("LN248 CBS");
                //if (node.colapse == 1)
                //    node.colapse = 0;
            }
            if (node.colapse != 2 && Math.Max(minCost, conflict.timeStep) > node.pathLength(conflict.agentB))
            {
                if (node.colapse != 0)
                    throw new Exception("expand/CBS");
                node.colapse = 2;
                node.totalCost += (conflict.timeStep + 1 - node.pathLength(conflict.agentB));
                node.totalLB += (conflict.timeStep + 1 - node.pathLength(conflict.agentB));
                openList.Add(node);
                if (node.totalCost <= threshold)
                    this.focalList.Add(new FocalNodeCBS(node));

                //Console.WriteLine("Colapse");
            }
            else if (!colapse || node.colapse == 2)
            {
                con = new DnCConstraint(conflict, instance, false);
                toAdd = new DnCNode(node, con, conflict.agentB, instance);

                if (closedList.Contains(toAdd) == false)
                {
                    if (toAdd.rePlan(instance, runner, conflict.agentB, Math.Max(minCost, getReplanDepth(node, conflict.agentB)), solver, lowLevelSolver, ref highLevelExpanded, ref highLevelGenerated, ref loweLevelExpanded, ref loweLevelGenerated))
                    {
                        if (toAdd.totalCost <= this.maxCost)
                        {
                            //if (toAdd.totalCost < prtCost)
                            //    Console.WriteLine("Stop!!");

                            //if (toAdd.totalCost < fMin)
                            //    Console.WriteLine("Stop!!");

                            //if (toAdd.totalCost <= fMin / 1.1)
                            //    Console.WriteLine("Stop!!");

                            //Console.WriteLine("GEN: " + toAdd.totalCost + "|" + toAdd.replanNum + "|" + toAdd.vCover + "|" + toAdd.couple + "|" + toAdd.internalConflictsCount + "|" + toAdd.costExp);

                            //foreach (SinglePlan plan in toAdd.allSingleAgentPlans)
                            //{
                            //    plan.PrintPlan();
                            //}

                            openList.Add(toAdd);
                            closedList.Add(toAdd);
                            this.highLevelGenerated++;
                            addToGlobalConflictCount(toAdd.getConflict());

                            if (toAdd.totalCost <= threshold)
                                this.focalList.Add(new FocalNodeCBS(toAdd));
                        }
                    }
                }
                //else
                //    Console.WriteLine("LN279 CBS");
                //if (node.colapse == 2)
                //    node.colapse = 0;
            }
            if (colapse)
                node.colapse = 0;
            return true;
        }
    }
}
