﻿/**
 * EH: For purposes of debugging and testing, I disabled the generation of the
 * Excel spreadsheet. You can turn it back on by defining (as true), EXCEL, in
 * this file as well as in Program.cs.
 */

//#define EXCEL

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;

namespace CPF_experement
{
    /// <summary>
    /// This class is responsible for running the experiments.
    /// </summary>
    public class Run
    {
        ////////debug
       // public static TextWriter resultsWriterdd;
        /////////////



        /// <summary>
        /// Delimiter character used when writing the results of the runs to the output file.
        /// </summary>
        public static string RESULTS_DELIMITER = ",";

        /// <summary>
        /// Number of random steps performed  when generating a new problem instance for choosing a start-goal pair.
        /// </summary>
        public static int RANDOM_WALK_STEPS = 100000;

        /// <summary>
        /// indecates the starting time in ms for timing the different algorithms.
        /// </summary>
        private long startTime;

        
        /// <summary>
        /// This hold an open stream to the results file.
        /// </summary>
        private TextWriter resultsWriter;

        /// <summary>
        /// EH: I introduced this variable so that debugging and experiments
        /// can have deterministic results.
        /// </summary>
        static public Random rand = new Random();

        /// <summary>
        /// Open the results file for output. Currently the file is opened in append mode.
        /// </summary>
        /// <param name="fileName">The name of the results file</param>
        public void openResultFile(string fileName)
        {
            this.resultsWriter = new StreamWriter(fileName,true); // 2nd arguments indicate the "append" mode

        }

        /// <summary>
        /// Closes the results file.
        /// </summary>
        public void closeResultsFile()
        {
            this.resultsWriter.Close();
        }

        /// <summary>
        /// all types of algorithms to be run
        /// </summary>
        List<ISolver> solvers;

        /// <summary>
        /// counts the number of times each algorithm went out of time
        /// </summary>
        public int[] outOfTimeCounter;

        /// <summary>
        /// cunstruct with chosen algorithms
        /// </summary>
        public Run()
        {
            solvers = new List<ISolver>();




         //   solvers.Add(new CostTreeSearchSolverNoPruning());
            //solvers.Add(new CostTreeSearchSolverKMatch(2));
            //solvers.Add(new CostTreeSearchSolverOldMatching(2));
            //solvers.Add(new CostTreeSearchSolverRepatedMatch(2));
            //solvers.Add(new CostTreeSearchSolverKMatch(3));
           // solvers.Add(new CostTreeSearchSolverOldMatching(3));
            //solvers.Add(new CostTreeSearchSolverRepatedMatch(3));


           // solvers.Add(new CostTreeSearchNoPruning());
            //solvers.Add(new CostTreeSearchKMatch(2));
            //solvers.Add(new CostTreeSearchOldMatching(2));
            //solvers.Add(new CostTreeSearchRepatedMatch(2));
            //solvers.Add(new CostTreeSearchKMatch(3));
    //        solvers.Add(new CostTreeSearchOldMatching(3));
            //solvers.Add(new CostTreeSearchRepatedMatch(3));

            //solvers.Add(new ClassicAStar());

           //solvers.Add(new AStarWithPartialExpansionBasic());
           // solvers.Add(new AStarWithOD());
           // solvers.Add(new AStarWithPartialExpansion());

   //         solvers.Add(new Trevor(new AStarWithPartialExpansion()));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 1, 1)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 5, 5)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 10, 10)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 100, 100)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 500, 500)));
   //         solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar())));
            //solvers.Add(new CostTreeSearchOldMatching(3));

            //solvers.Add(new CBS_IDA(new ClassicAStar()));
           // solvers.Add(new AStarWithPartialExpansion());
          //  solvers.Add(new CBS_GlobalConflicts(new ClassicAStar()));

           // solvers.Add(new CBS_NoDD(new ClassicAStar()));
            //solvers.Add(new CBS_NoDDb3(new ClassicAStar()));

           // solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar(), 1, 1)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar(), 5, 5)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar(), 10, 10)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar(), 100, 100)));
            //solvers.Add(new Trevor(new CBS_GlobalConflicts(new ClassicAStar(), 500, 500)));


            //solvers.Add(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 1, 1));
            //solvers.Add(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 5, 5));
            //solvers.Add(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 10, 10));
            //solvers.Add(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 100, 100));
           // solvers.Add(new CBS_GlobalConflicts(new AStarWithPartialExpansion(), 500, 500));
           // solvers.Add(new CBS_GlobalConflicts(new ClassicAStar()));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStar(), 0, 0));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStar(), 1, 0));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStar(), 2, 0));
            
            //if(Constants.H_VER==0)
                //solvers.Add(new CBS_LocalConflicts(new ClassicAStar()));

            //solvers.Add(new ClassicAStarE());

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.1)));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.5)));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(2)));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(100)));

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(100)));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.6));

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStar()));
            //solvers.Add(new CBSE(new ClassicAStarE(1.1),1.1));

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.1)));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.1));
            //solvers.Add(new CBSE(new ClassicAStarE(1.1), 1.1));
            

            //solvers.Add(new ClassicAStar());

            //solvers.Add(new CBSE(new ClassicAStar(), 1.5));

            //solvers.Add(new CBSE(new ClassicAStar(), 1.7));

            //solvers.Add(new CBSE(new ClassicAStar(), 1.9));

            //solvers.Add(new CBSE(new ClassicAStar(), 2));

            //solvers.Add(new CBSE(new ClassicAStar(), 2.5));

            //solvers.Add(new CBSE(new ClassicAStar(), 1.8));
            
            //solvers.Add(new CBSE(new ClassicAStar(), 2));

            //if (Constants.H_VER != 0)
             //   solvers.Add(new CBSE(new ClassicAStar(), 100));

            //solvers.Add(new CBSE(new ClassicAStar(), 1.1));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.5));
            //solvers.Add(new CBSE(new ClassicAStar(), 100));

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.1)));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.5)));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(100)));

            ////solvers.Add(new CBS_LocalConflicts(new ClassicAStar()));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.1)));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.1));
            //solvers.Add(new CBSE(new ClassicAStarE(1.048), 1.048));

            //solvers.Add(new CBS_LocalConflicts(new ClassicAStar()));
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(100)));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.5));
            //solvers.Add(new CBSE(new ClassicAStarE(100), 100));


            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.01)));
            //solvers.Add(new CBSE(new ClassicAStarE(1.35), 1.1));
            //solvers.Add(new CBSE(new ClassicAStarE(1.00498), 1.00498));
            //solvers.Add(new CBSE(new ClassicAStarE(1.11), 1.35));
            //solvers.Add(new CBSE(new ClassicAStar(), 1.01));

            //GCBS-L
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(100)));

            //GCBS-H
            //solvers.Add(new CBSE(new ClassicAStar(), 100));

            //GCBS-LH
            //solvers.Add(new CBSE(new ClassicAStarE(100), 100));

            //BCBS-L(1.5)
            //solvers.Add(new CBS_LocalConflicts(new ClassicAStarE(1.5)));

            //BCBS-H(1.5)
            //solvers.Add(new CBSE(new ClassicAStar(), 1.5));

            //BCBS-LH(1.5)
            //solvers.Add(new CBSE(new ClassicAStarE(1.22), 1.22));

            //ECBS(1.5)
            //solvers.Add(new ECBSE(new ClassicAStarE(1.5), 1.5));
            //solvers.Add(new ECBSE(new ClassicAStarE(1.1), 1.1));
            //solvers.Add(new ECBSE(new ClassicAStarE(1.05), 1.05));
            //solvers.Add(new ECBSE(new ClassicAStarE(1.01), 1.01));

            // smooth
            for (double e = 1.1; e >= 1; e -= 0.01)
            {
                solvers.Add(new ECBSE(new ClassicAStarE(e), e));
            }

            //solvers.Add(new ECBSE(new ClassicAStarE(1.01), 1.01));

            //solvers.Add(new ECBSE(new ClassicAStarE(1.5), 1.5));
            //solvers.Add(new ECBSE(new ClassicAStarE(2), 2));
            //solvers.Add(new CBSE(new ClassicAStarE(2), 2));

            //solvers.Add(new Trevor(new AStarWithOD()));

           // solvers.Add(new Trevor(new AStarWithPartialExpansionBasic()));
           // solvers.Add(new Trevor(new AStarWithPartialExpansion()));
           // solvers.Add(new Trevor(new ClassicAStar()));
           // solvers.Add(new Trevor());

            outOfTimeCounter = new int[solvers.Count];
            for (int i = 0; i < outOfTimeCounter.Length; i++)
            {
                outOfTimeCounter[i] = 0;
            }
        }

        /// <summary>
        /// Generates a problem instance, including a board, start and goal locations of desired number of agents
        /// and desired precentage of obstacles
        ///  TODO: Refactor to use operators.
        /// </summary>
        /// <param name="gridSize"></param>
        /// <param name="agentsNum"></param>
        /// <param name="obsticalsNum"></param>
        /// <returns></returns>
        public ProblemInstance generateProblemInstance(int gridSize, int agentsNum, int obsticalsNum)
        {
            /**
             * Randomization based on timer is disabled for purposes of getting
             * reproducible experiments.
             */
            //Random rand = new Random();

            int x;
            int y;
            Agent[] aGoals = new Agent[agentsNum];
            AgentState[] aStart = new AgentState[agentsNum];
            bool[][] grid = new bool[gridSize][];
            bool[][] goals = new bool[gridSize][];

            // Generate a random grid
            for (int i = 0; i < gridSize; i++)
            {
                grid[i] = new bool[gridSize];
                goals[i] = new bool[gridSize];
            }
            for (int i = 0; i < obsticalsNum; i++)
            {
                x = rand.Next(gridSize);
                y = rand.Next(gridSize);
                if (grid[x][y])
                    i--;
                grid[x][y] = true;
            }

            // Choose random goal locations
            for (int i = 0; i < agentsNum; i++)
            {
                x = rand.Next(gridSize);
                y = rand.Next(gridSize);
                if (goals[x][y] || grid[x][y])
                    i--;
                else
                {
                    goals[x][y] = true;
                    aGoals[i] = new Agent(x, y, i);
                }
            }

            // Select random start/goal locations for every agent by performing a random walk
            for (int i = 0; i < agentsNum; i++)
            {
                aStart[i] = new AgentState(aGoals[i].Goal_X, aGoals[i].Goal_Y, aGoals[i]);
            }

            // Initialzied here only for the IsValidForMove() call. TODO: Think how this can be sidestepped elegantly.
            ProblemInstance problem = new ProblemInstance();
            problem.init(aStart, grid);
            
            int op;
            int deltaX, deltaY,newX,newY;
            for (int j = 0; j < RANDOM_WALK_STEPS; j++)
            {
                for (int i = 0; i < agentsNum; i++)
                {
                    op = rand.Next(0, WorldState.operators.GetLength(0));
                    deltaX = WorldState.operators[op, 0];
                    deltaY = WorldState.operators[op, 1];
                    newX = aStart[i].pos_X + deltaX;
                    newY = aStart[i].pos_Y + deltaY;
                    if (problem.IsValidForMove(newX, newY) && !goals[newX][newY])
                    {
                        goals[newX][newY] = true;
                        goals[aStart[i].pos_X][aStart[i].pos_Y] = false;
                        aStart[i].pos_X += deltaX;
                        aStart[i].pos_Y += deltaY;
                    }
                }
            }

            // TODO: There is some repetetition here of previous instansiation of ProblemInstance. Think how to elegantly bypass this.
            problem = new ProblemInstance();
            problem.init(aStart, grid);
            return problem;            
        }


        /// <summary>
        /// Solve given instance with a list of algorithms 
        /// </summary>
        /// <param name="instance">The instance to solve</param>
        public void solveGivenProblem(ProblemInstance instance)
        {
            Debug.WriteLine("Solving instance " + instance);
            int gridSize = instance.m_vGrid.Length;
            //this.printProblemStatistics(instance);
            CBS_LocalConflicts.isDnC = true;
            //double cr0 = instance.getConflictRation(0);
            //double cr1 = instance.getConflictRation(1);
            CBS_LocalConflicts.isDnC = false;

            //Debug.WriteLine("Conflict ratio (first order): " + cr0 );
            //Debug.WriteLine("Conflict ratio (second order): " + cr1);
            //this.resultsWriter.Write(cr0 + RESULTS_DELIMITER);
            //this.resultsWriter.Write(cr1 + RESULTS_DELIMITER);


            Constants.exausitveIcts = false;

            for (int i = 0; i < solvers.Count; i++)
            {
                //if (i == 1)
                //    Constants.exausitveIcts = true;

                if (outOfTimeCounter[i] < Constants.MAX_FAIL_COUNT)//after "MAX_FAIL_COUNT" failures of a given algotrithm we stop running it
                {
                    GC.Collect();
                    GC.WaitForPendingFinalizers();
                    this.printProblemStatistics(instance);
                    this.run(solvers[i], instance);
                    this.resultsWriter.WriteLine();

                    //solvers[i].GetPlan().PrintPlan();

                    Console.WriteLine();
                    if (solvers[i].GetSolutionCost() > 0)
                    {
                        outOfTimeCounter[i] = 0;
                        Console.WriteLine("+SUCCESS+ (:");
                        //  solvers[i].GetPlan().PrintPlan();
                        if (solvers[0].GetSolutionCost() > 0)
                        {
                            //Debug.Assert(solvers[0].GetSolutionCost() == solvers[i].GetSolutionCost(), "A* solution cost is different than that of " + solvers[i].GetName());
                            //Debug.Assert(solvers[0].getExpanded() == solvers[i].getExpanded(), "Different Expended");
                           // Debug.Assert(solvers[0].getGenerated() == solvers[i].getGenerated(), "Different Generated");

                            //if (solvers[0].GetSolutionCost() != solvers[i].GetSolutionCost())
                            //{
                            //    Console.WriteLine("A* solution cost is different than that of " + solvers[i].GetName());
                            //    Console.ReadLine();
                            //}
                           // Debug.Assert(solvers[0].getSolutionDepth() == solvers[i].getSolutionDepth(), "Depth Bug " + solvers[i].GetName());
                            //ss = new ShowSolution(instance, solvers[i].GetPlan());
                            //ss.Show();
                        }
                    }
                    else
                    {
                        outOfTimeCounter[i]++;
                        Console.WriteLine("-FAILURE- ):");
                    }
                }
                else
                    printNullStatistics();
                
                Console.WriteLine();
            }
            //this.resultsWriter.WriteLine();
            this.resultsWriter.Flush();
          //  Debug.Assert(solvers[0].getNodesPassedPruningCounter() >= solvers[1].getNodesPassedPruningCounter(), "");
           // Console.ReadLine();
        }

        /// <summary>
        /// Solve a given instance with the given solver
        /// </summary>
        /// <param name="solver">The solver</param>
        /// <param name="instance">The problem instance that will be solved</param>
        private void run(ISolver solver, ProblemInstance instance)
        {
            bool solved;
            Console.WriteLine("-----------------" + solver.GetName() + "-----------------");
            this.startTime = this.ElapsedMillisecondsTotal();
            solver.Setup(instance);
            solved = solver.Solve(this);
            this.startTime = this.ElapsedMilliseconds();
            if (solved)
            {
                Console.WriteLine("Total cost: {0}", solver.GetSolutionCost());
            }
            else
            {
                Console.WriteLine("Failed to solve");
            }
            Console.WriteLine();
            Console.WriteLine("Time In milliseconds: {0}", this.startTime);
            Console.WriteLine("Total Expanded Nodes (High-Level): {0}", solver.getHighLeveExpanded());
            Console.WriteLine("Total Generated Nodes (High-Level): {0}", solver.getHighLeveGenerated());
            Console.WriteLine("Total Expanded Nodes (Low-Level): {0}", solver.getLowLevelExpanded());
            Console.WriteLine("Total Generated Nodes (Low-Level): {0}", solver.getLowLevelGenerated());
           // Console.WriteLine("Total Unique/Full Expanded Nodes: {0}", solver.getNodesPassedPruningCounter());

            this.printStatistics(instance, solver, this.startTime);
            solver.Clear();
        }

        /// <summary>
        /// Print the header of the results file
        /// </summary>
        public void printResultsFileHeader()
        {
            this.resultsWriter.Write("GridSize");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("NumOfAgents");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("NumOfObstacles");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("InstanceId");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            //this.resultsWriter.Write("Conflict1");
            //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            //this.resultsWriter.Write("Conflict2");
            //this.resultsWriter.Write(Run.RESULTS_DELIMITER);

            //for (int i = 0; i < solvers.Count; i++)
            //{
                //this.resultsWriter.Write(solvers[i].GetName() + "success");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Runtime");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "SolutionCost");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Expanded-HL");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Generated-HL");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Expanded-LL");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Generated-LL");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Max Group");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);


                //this.resultsWriter.Write(solvers[i].GetName() + "Min Group / G&D");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Max depth");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Passed Nodes/Expanded Full States");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
                //this.resultsWriter.Write(solvers[i].GetName() + "Memory Used");
                //this.resultsWriter.Write(Run.RESULTS_DELIMITER);
           // }

            this.resultsWriter.Write("success");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Runtime");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("SolutionCost");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Expanded-HL");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Generated-HL");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Expanded-LL");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Generated-LL");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Max Group");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Algorithm");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);
            this.resultsWriter.Write("Heuristic");
            this.resultsWriter.Write(Run.RESULTS_DELIMITER);


            this.resultsWriter.WriteLine();
        }


        /// <summary>
        /// Print the solver statistics to the results file.
        /// </summary>
        /// <param name="instance">The problem instance that was solved</param>
        /// <param name="solver">The solver that solved the problem instance</param>
        /// <param name="runtimeInMillis">The time it took the given solver to solve the given instance</param>
        private void printStatistics(ProblemInstance instance, ISolver solver, long runtimeInMillis)
        {
            if (solver.GetSolutionCost() < 0)
            {
                this.resultsWriter.Write(0 + RESULTS_DELIMITER);
            }
            else
            {
                this.resultsWriter.Write(1 + RESULTS_DELIMITER);
            }

            this.resultsWriter.Write(runtimeInMillis + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.GetSolutionCost() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.getHighLeveExpanded() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.getHighLeveGenerated() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.getLowLevelExpanded() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.getLowLevelGenerated() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.getMaxGroupSize() + RESULTS_DELIMITER);
            this.resultsWriter.Write(solver.GetName() + RESULTS_DELIMITER);
            this.resultsWriter.Write(Constants.VER_DESC[Constants.H_VER] + RESULTS_DELIMITER);

            

            //solver.OutputStatistics(this.resultsWriter);
        }
        private void printProblemStatistics(ProblemInstance instance)
        {
            this.resultsWriter.Write(instance.m_vGrid.GetLength(0) + RESULTS_DELIMITER);
            this.resultsWriter.Write(instance.m_vAgents.Length + RESULTS_DELIMITER);
            this.resultsWriter.Write(instance.m_nObstacles + RESULTS_DELIMITER);
            this.resultsWriter.Write(instance.instanceId + RESULTS_DELIMITER);
        }
        private void continueToNextLine()
        {
            this.resultsWriter.WriteLine();
            this.resultsWriter.Flush();
        }
        private void printNullStatistics()
        {
            this.resultsWriter.Write(0 + RESULTS_DELIMITER);
            this.resultsWriter.Write(Constants.MAX_TIME + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
            this.resultsWriter.Write("irrelevant" + RESULTS_DELIMITER);
        }
        public void resetOutOfTimeCounter()
        {
            for (int i = 0; i < outOfTimeCounter.Length; i++)
            {
                outOfTimeCounter[i] = 0;
            }
        }

        //public void setOutOfTimeCounter()
        //{
        //    Console.WriteLine("Enter Faile Count Out Of Total - " + Constants.MAX_FAIL_COUNT);
        //    Console.WriteLine("---------------------------------------");
        //    for (int i = 0; i < solvers.Count; i++)
        //    {
        //        Console.WriteLine("Enter Fail Count For " + solvers[i].GetName());
        //        this.outOfTimeCounter[i] = Int16.Parse(Console.ReadLine());
        //    }
        //}

        private long ElapsedMillisecondsTotal()
        {
            TimeSpan interval = Process.GetCurrentProcess().TotalProcessorTime;
            long nMilliseconds = interval.Days * 24 * 60 * 60 * 1000 +
                           interval.Hours * 60 * 60 * 1000 +
                           interval.Minutes * 60 * 1000 +
                           interval.Seconds * 1000 +
                           interval.Milliseconds;
            return nMilliseconds;
        }

        public long ElapsedMilliseconds()
        {
            return ElapsedMillisecondsTotal() - this.startTime;
        }

       

    }
}
